unit Netz;

interface
uses
  Dialogs,SysUtils, Math;

type
  Net = class
private
  HFNetz: Array of Array of Integer;
  FundMust: Array of Array of Array of Integer;
  w: Array of Array of Real;
  Neuronenanzahl: Integer;
  theta: Array of Real;
  AnzahlPropagierungsschritte: Integer;
public
  constructor create(x,y: Integer);
  function toggle(x,y: INteger):Integer;
  function toString(): String;
  procedure initFundMuster(Anzahl: Integer);
  function FundMusttoString(FundMustNr: Integer): String;
  procedure setzeFundMust(FundMustNr,x,y,wert:Integer);
  procedure uebernehmeAlsFundMust(FundMusterNr:Integer);
  function highx:Integer;
  function highy:Integer;
  function getFundMust(Nr,x,y:Integer):Integer;
  procedure fundMusttoAkt(Nr: Integer);
  function getw(i,j: Integer):Real;
  procedure setw(i,j: Integer; wert: Real);
  function getxPosOfNr(nr: Integer): Integer;
  function getyPosOfNr(nr: Integer): Integer;
  function getNrOfPos(x,y: Integer): Integer;
  procedure learn;
  procedure setTheta(nr: Integer; wert: Real); overload; // ein Neuron
  procedure setTheta(wert: Real); overload; // alle Neuronen
  function getNeuronenanzahl: Integer;
  function wToString: String;
  procedure PropAlgSchritt;
  procedure Propagiere;
  function BerechneNeuron(Nr: Integer): Integer;
  function getAusgabe(x,y: Integer):Integer;
  function getHighFundMust: Integer;
  function getAnzahlpropagierungsschritte: Integer;
  procedure setAnzahlpropagierungsschritte(i: Integer);
  procedure setField(x,y,i:Integer);
end;

implementation

constructor Net.create(x,y: Integer);
var
  i,j:Integer;
begin
// Initialisiere Matrix
setLength(HFNetz, x);
for i:=0 to high(HFNetz) do
  setLength(HFNetz[i], y);
for i:=0 to high(HFNetz) do
  for j:=0 to high(HFNetz[0]) do
    HFNetz[i,j]:=-1;

// Setze Neuronenanzahl
  Neuronenanzahl:=x*y;

// Initialisiere Gewichtsmatrix w
setlength(w, Neuronenanzahl);
for i:=0 to high(w) do
  setLength(w[i], Neuronenanzahl);
for i:=0 to high(w) do
  for j:=0 to high(w[0]) do
    w[i,j]:=0;

// Initialisiere Schwellwertarray theta
setlength(theta, Neuronenanzahl);
for i:=0 to high(theta) do
  theta[i]:=0;

// Anzahl Propagierungsschritte erstmal 10 * Neuronenanzahl
AnzahlPropagierungsschritte := 10;
end;

function Net.toggle(x,y: Integer):Integer;
begin
x:=x-1;
y:=y-1;
case HFNetz[x,y] of
  -1: HFNetz[x,y]:=  1;
   1: HFNetz[x,y]:= -1;
end;
result:=HFNetz[x,y];
end;

function Net.toString(): String;
var
  str: String;
  i,j: Integer;
begin
  str:='';
  for j:=0 to High(HFNetz[0]) do
  begin
    for i:=0 to High(HFNetz) do
      str:=str + ' ' + IntToStr(HFNetz[i,j]);
    str:=str+#13+#10;
  end;
  result:=str;
end;

procedure Net.initFundMuster(Anzahl: Integer);
var i,j,k: Integer;
begin
setlength(FundMust, Anzahl);
for i:=0 to High(FundMust) do
begin
  setlength(FundMust[i], length(HFNetz));
  for j:=0 to High(FundMust[i]) do
  begin
    setlength(FundMust[i,j], length(HFNetz[0]));
  end;
end;

// Mit -1 initialisieren
for i:=0 to High(FundMust) do
  for j:=0 to High(FundMust[i]) do
    for k:=0 to High(FundMust[i,j]) do
      FundMust[i,j,k]:=-1;
end;

function Net.FundMusttoString(FundMustNr: Integer): String;
var
  str: String;
  i,j: Integer;
begin
  str:='';
  for j:=0 to High(FundMust[FundMustNr,0]) do
  begin
    for i:=0 to High(FundMust[FundMustNr]) do
      str:=str + ' ' + IntToStr(FundMust[FundMustNr,i,j]);
    str:=str+#13+#10;
  end;
  result:=str;
end;

procedure Net.setzeFundMust(FundMustNr,x,y,wert:Integer);
begin
  FundMust[FundMustNr,x,y]:=wert;
end;

procedure Net.uebernehmeAlsFundMust(FundMusterNr:Integer);
var i,j: Integer;
begin
  for i:=0 to High(HFNetz) do
    for j:=0 to High(HFNetz[0]) do
      FundMust[FundMusterNr,i,j]:=HFNetz[i,j];
end;

function Net.highx:Integer;
begin
  result:=high(HFNetz);
end;

function Net.highy:Integer;
begin
  result:=high(HFNetz[0]);
end;

function Net.getFundMust(Nr,x,y:Integer):Integer;
begin
  result:=FundMust[Nr,x,y];
end;

procedure Net.fundMusttoAkt(Nr: Integer);
var a,b:Integer;
begin
  for a:=0 to Highx do
    for b:=0 to Highy do
      HFNetz[a,b]:=fundMust[Nr,a,b]
end;

function Net.getw(i,j: Integer): Real;
begin
// symmetrische Matrix: Ich nehme "immer erst den kleineren Index"
if i<j then result:=w[i,j]
else if j<i then result:=w[j,i]
else result:=0;
end;

procedure Net.setw(i,j: Integer; wert: Real);
begin
// symmetrische Matrix: Ich nehme "immer erst den kleineren Index"
if i<j then w[i,j]:=wert
else if j<i then w[j,i]:=wert;
end;

function Net.getxPosOfNr(nr: Integer): Integer;
begin
  result := nr mod length(HFNetz);
end;

function Net.getyPosOfNr(nr: Integer): Integer;
begin
  setRoundMode(rmDown);
  result := round(nr / length(HFNetz));
end;

function Net.getNrOfPos(x,y: Integer): Integer;
begin
  result := y*length(HFNetz)+x;
end;

procedure Net.learn;
var
  i,j,mue:integer;
  sum: real;
begin
  // fr jede Verbindung...
  for i:=0 to Neuronenanzahl-1 do
    for j:=0 to Neuronenanzahl-1 do
      begin
        // fr jedes Fundamentalmuster mue nehme ich den des Anfangs und des Endes der Verbindung
        sum:=0;
        for mue:=0 to High(FundMust) do
          sum := sum + FundMust[mue,getxPosOfNr(i),getyPosOfNr(i)] * FundMust[mue,getxPosOfNr(j),getyPosOfNr(j)];
        sum:= (sum / Neuronenanzahl);
        // und Weise ihn dem Gewicht w_ij zu
        setw(i,j,sum);
      end;
end;

procedure Net.setTheta(nr: Integer; wert: Real); // ein Neuron
begin
  Theta[nr]:=wert;
end;

procedure Net.setTheta(wert: Real); // alle Neuronen
var i: Integer;
begin
for i:=0 to high(theta) do
  theta[i]:=wert;
end;

function Net.getNeuronenanzahl: Integer;
begin
result:=Neuronenanzahl;
end;

function Net.wToString: String;
var
  str: String;
  i,j: Integer;
begin
  str:='';
  for j:=0 to High(w[0]) do
  begin
    for i:=0 to High(w) do
      str:=str + ' ' + FloatToStr(w[i,j]);
    str:=str+#13+#10;
  end;
  result:=str;
end;

procedure Net.PropAlgSchritt;
var
  aktNeuronNr: Integer;
begin
  aktNeuronNr := RandomRange(0,Neuronenanzahl);
  BerechneNeuron(aktNeuronNr);
end;

procedure Net.Propagiere;
var q: Integer;
begin
  for q:=0 to Anzahlpropagierungsschritte * Neuronenanzahl do
    PropAlgSchritt;
end;

function Net.BerechneNeuron(Nr: Integer): Integer;
var
  sum: Real;
  i: Integer;
begin
  // Interner Zustand
    // Gewichtete Summe ber alle Eingnge
    sum:=0;
    for i:=0 to Neuronenanzahl-1 do
      sum := sum + getw(Nr, i) * HFNetz[getxPosOfNr(i), getyPosOfNr(i)];
    sum := sum - theta[Nr];
    // Aktivierungsfunktion sgn*
    if      sum > 0 then HFNetz[getxPosOfNr(Nr), getyPosOfNr(Nr)]:= 1
    else if sum < 0 then HFNetz[getxPosOfNr(Nr), getyPosOfNr(Nr)]:= -1;
    //Showmessage('Koordinaten: '+inttostr(getxPosOfNr(Nr))+'; '+inttostr(getyPosOfNr(Nr)));
    result := HFNetz[getxPosOfNr(Nr), getyPosOfNr(Nr)];
end;

function Net.getAusgabe(x,y: Integer):Integer;
begin
  result:=HFNetz[x,y];
end;

function Net.getHighFundMust: Integer;
begin
  result:= High(FundMust);
end;

function Net.getAnzahlpropagierungsschritte: Integer;
begin
  result:=Anzahlpropagierungsschritte;
end;

procedure Net.setAnzahlpropagierungsschritte(i: Integer);
begin
  Anzahlpropagierungsschritte:=i;
end;

procedure Net.setField(x,y,i:Integer);
begin
  HFNetz[x,y]:=i;
end;

end.

